/* args.c - Parse command-line argument for xsvg using argp
 *
 * Copyright  2002 USC/Information Sciences Institute
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Information Sciences Institute not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  Information Sciences Institute
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * INFORMATION SCIENCES INSTITUTE DISCLAIMS ALL WARRANTIES WITH REGARD
 * TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL INFORMATION SCIENCES
 * INSTITUTE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl Worth <cworth@isi.edu>
 */

#include <stdlib.h>
#include <argp.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "args.h"

const char *argp_program_version = "xsvg " VERSION;

const char *argp_program_bug_address = "<cworth@isi.edu>";

static char doc[] = "xsvg - X11 SVG viewer";

static char args_doc[] = "<SVG file>";

static struct argp_option options[] = {
    /* name,		key, arg,	 flags, doc */
    {"display",		'd', "DISPLAY",  0, "X server to contact" },
    {"geometry",	'g', "GEOMETRY", 0, "Size and location of window" },
    {"scale",		's', "FACTOR",   0, "Scale image by FACTOR" },
    {"png",		'p', "FILENAME", 0, "Render as PNG to FILENAME instead of a window (EXPERIMENTAL)" }, 
    {"width",		'w', "WIDTH",	 0, "Width for output PNG"},
    {"height",		'h', "HEIGHT",	 0, "Height for output PNG"},
    {"flipx",		'x', 0,		 0, "Flip X coordinates"},
    {"flipy",		'y', 0,		 0, "Flip Y coordinates"},
    { 0 }
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
    struct args *args = state->input;

    switch (key) {
    case 'd':
	args->display = arg;
	break;
    case 'g':
	args->geometry = arg;
	break;
    case 'p':
	args->png_file = arg;
	break;
    case 's':
	args->scale = atof(arg);
	break;
    case 'w':
	args->width = atoi(arg);
	break;
    case 'h':
	args->height = atoi(arg);
	break;
    case 'x':
	args->flipx = 1;
	break;
    case 'y':
	args->flipy = 1;
	break;

    case ARGP_KEY_ARG:
	if (state->arg_num > 0)
	    argp_usage (state);

	args->svg_file = arg;

	break;

    case ARGP_KEY_END:
	if (state->arg_num < 1)
	    /* Didn't see a file option */
	    argp_usage (state);
	break;

    default:
	return ARGP_ERR_UNKNOWN;
    }

    return 0;
}
	    
static struct argp argp = { options, parse_opt, args_doc, doc };

error_t
args_parse(args_t *args, int argc, char *argv[])
{
    args->display = NULL;
    args->geometry = NULL;
    args->png_file = NULL;
    args->scale = 1.0;

    /* XXX: The whole width/height args thing is pretty bogus. xsvg
       should be reading size information from the SVG file. */
    args->width = 450;
    args->height = 450;
    args->flipx = 0;
    args->flipy = 0;

    return argp_parse (&argp, argc, argv,
		       ARGP_LONG_ONLY,
		       NULL, args);
}
