/* xsvg - SVG viewer application for the X Window System
 *
 * Copyright  2002 USC/Information Sciences Institute
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Information Sciences Institute not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  Information Sciences Institute
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * INFORMATION SCIENCES INSTITUTE DISCLAIMS ALL WARRANTIES WITH REGARD
 * TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL INFORMATION SCIENCES
 * INSTITUTE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl Worth <cworth@isi.edu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <png.h>

#include <xsvg.h>

#include "args.h"

#define SHIFT 10

#define ARRAY_SIZE(A) (sizeof(A)/sizeof(A[0]))

typedef struct win {
    Display *dpy;
    int scr;
    Window win;
    unsigned long event_mask;
    Pixmap pix;
    GC gc;

    unsigned int width, height;
    int x_flip, y_flip;

    cairo_t *xrs;
    int needs_refresh;

    xsvg_t *xsvg;

    double tx;
    double ty;
    double x_zoom;
    double y_zoom;
    double tolerance;
} win_t;

typedef struct callback_doc {
    void *callback;
    char *doc;
} callback_doc_t;

typedef int (*key_callback_t)(win_t *win);

typedef struct key_binding
{
    char *key;
    int is_alias;
    KeyCode keycode;
    key_callback_t callback;
} key_binding_t;

static void
win_init (win_t *win, Display *dpy, char *geometry, const char *svg_file);

static void
win_deinit (win_t *win);

static void
win_refresh (win_t *win);

static void
win_grow_pixmap(win_t *win);

static int
win_handle_key_press(win_t *win, XKeyEvent *kev);

static void
win_reconfigure (win_t *win, unsigned int width, unsigned int height);

static void
win_handle_configure (win_t *win, XConfigureEvent *cev);

static void
win_handle_expose(win_t *win, XExposeEvent *eev);

static void
win_handle_events(win_t *win);

static void
render_to_png (const char *svg_file, const char *png_file, double scale, int width, int height);

static svg_status_t
write_png_argb32 (char *buffer, const char* filename,
		  int width, int height, int stride);

/* callbacks */
static int
quit_cb(win_t *win);
static int
x_flip_cb(win_t *win);
static int
y_flip_cb(win_t *win);
static int
left_cb(win_t *win);
static int
right_cb(win_t *win);
static int
up_cb(win_t *win);
static int
down_cb(win_t *win);
static int
zoom_in_cb(win_t *win);
static int
zoom_out_cb(win_t *win);
static int
zoom_x_cb(win_t *win);
static int
zoom_y_cb(win_t *win);
static int
flatten_cb(win_t *win);
static int
smooth_cb(win_t *win);

static const callback_doc_t callback_doc[] = {
    { quit_cb,		"Exit the program" },
    { zoom_in_cb,	"Zoom in  (1.1X)" },
    { zoom_out_cb,	"Zoom out (1.1X)" },
    { x_flip_cb,	"Flip by inverting X values." },
    { y_flip_cb,	"Flip by inverting Y values." },
    { smooth_cb,	"Increase rendering accuracy (10X)" },
    { flatten_cb,	"Decrease rendering accuracy (10X)" },
};

static key_binding_t key_binding[] = {
    /* Keysym, Alias, Keycode, callback */
    { "X",	0, 0, zoom_x_cb },
    { "Y",	0, 0, zoom_y_cb },
    { "Q",	0, 0, quit_cb },
    { "Left",	0, 0, left_cb },
    { "Right",	0, 0, right_cb },
    { "Up",	0, 0, up_cb },
    { "Down",	0, 0, down_cb },
    { "plus",	0, 0, zoom_in_cb },
    { "minus",	0, 0, zoom_out_cb },
    { "greater",0, 0, smooth_cb },
    { "less",	0, 0, flatten_cb },
};

int 
main (int argc, char **argv)
{
    args_t args;
    win_t win;
    args_parse (&args, argc, argv);

    if (args.png_file) {
	render_to_png (args.svg_file, args.png_file, args.scale, args.width, args.height);
    } else {
	Display *dpy;

	dpy = XOpenDisplay (args.display);
	if (dpy == NULL) {
	    fprintf (stderr, "Failed to open display: %s\n", XDisplayName(args.display));
	    return 1;
	}

	win_init (&win, dpy, args.geometry, args.svg_file);
	win.x_zoom *= args.scale;
	win.y_zoom *= args.scale;
	win.x_flip = args.flipx;
	win.y_flip = args.flipy;

	win_handle_events (&win);

	win_deinit (&win);

	XCloseDisplay (dpy);
    }

    return 0;
}

/* XXX: This function needs to be fixed to use the actual image size
   from the SVG file rather than the ugly hard-coded value. We'll also
   want the ability to let the user specify a viewbox on the
   command-line. */
static void
render_to_png (const char *svg_file, const char *png_file, double scale, int width, int height)
{
    xsvg_status_t status;
    cairo_t *xrs;
    xsvg_t *xsvg;
    char *image;

    image = calloc (width * height, 4);
    if (! image)
	return;

    xrs = cairo_create ();
    cairo_scale (xrs, scale, scale);

    cairo_set_target_image (xrs, image, CAIRO_FORMAT_ARGB32, width, height, width * 4);

    /* XXX: This probably doesn't need to be here (eventually) */
    cairo_set_rgb_color (xrs, 1, 1, 1);

    status = xsvg_create (&xsvg);
    if (status) {
	fprintf (stderr, "Failed to create xsvg_t. Exiting.\n");
	exit(1);
    }

    status = xsvg_parse_file (xsvg, svg_file);
    if (status) {
	fprintf (stderr, "Failed to parse SVG file: %s. Exiting.\n", svg_file);
	exit(1);
    }

    xsvg_render (xsvg, xrs);

    write_png_argb32 (image, png_file, width, height, width * 4);

    xsvg_destroy (xsvg);
    xsvg = NULL;

    free (image);
}

static void
win_init (win_t *win, Display *dpy, char *geometry, const char *svg_file)
{
    unsigned int i;
    xsvg_status_t status;
    XGCValues gcv;

    win->dpy = dpy;
    win->scr = DefaultScreen (dpy);

    win->tx = 0;
    win->ty = 0;
    win->x_flip = 0;
    win->y_flip = 0;
    win->x_zoom = 1.0;
    win->y_zoom = 1.0;
    win->tolerance = 0.1;

    win->needs_refresh = 1;

    if (geometry) {
	int x, y;
	XParseGeometry (geometry, &x, &y, &win->width, &win->height);
    } else {
	/* XXX: Shall I get the window size from the SVG file perhaps */
	win->width = 450;
	win->height = 450;
    }

    win->win = XCreateSimpleWindow (dpy, DefaultRootWindow (dpy), 0, 0,
				    win->width, win->height, 0,
				    BlackPixel (dpy, win->scr), WhitePixel (dpy, win->scr));

    win->pix = XCreatePixmap(dpy, win->win, win->width, win->height, DefaultDepth(dpy, win->scr));
    gcv.foreground = WhitePixel(dpy, win->scr);
    win->gc = XCreateGC(dpy, win->pix, GCForeground, &gcv);
    XFillRectangle(dpy, win->pix, win->gc, 0, 0, win->width, win->height);

    for (i=0; i < ARRAY_SIZE(key_binding); i++) {
	KeySym keysym;
	keysym = XStringToKeysym(key_binding[i].key);
	if (keysym == NoSymbol)
	    fprintf(stderr, "ERROR: No keysym for \"%s\"\n", key_binding[i].key);
	else
	    key_binding[i].keycode = XKeysymToKeycode(dpy, keysym);
    }

    win->xrs = cairo_create ();
    cairo_set_target_drawable (win->xrs, dpy, win->pix);
    /* XXX: This probably doesn't need to be here (eventually) */
    cairo_set_rgb_color (win->xrs, 1, 1, 1);

    status = xsvg_create (&win->xsvg);
    if (status) {
	fprintf (stderr, "Failed to create xsvg_t. Exiting.\n");
	exit(1);
    }

    xsvg_set_viewport_dimension (win->xsvg, win->width, win->height);

    status = xsvg_parse_file (win->xsvg, svg_file);
    if (status) {
	fprintf (stderr, "Failed to parse SVG file: %s. Exiting.\n", svg_file);
	exit(1);
    }

    win->event_mask = (KeyPressMask
		       | StructureNotifyMask
		       | ExposureMask);

    XSelectInput (dpy, win->win, win->event_mask);

    XMapWindow (dpy, win->win);
}

static void
win_deinit (win_t *win)
{
    xsvg_destroy (win->xsvg);
    win->xsvg = NULL;
    cairo_destroy (win->xrs);
    win->xrs = NULL;
    XFreeGC (win->dpy, win->gc);
    XDestroyWindow (win->dpy, win->win);
}

static void
win_refresh (win_t *win)
{
/* Xr does not currently support statistics reporting

    struct timeval start_tv, end_tv;
    double elapsed;
    double start_traps, end_traps;
*/

    XFillRectangle (win->dpy, win->pix, win->gc, 0, 0, win->width, win->height);

    cairo_save (win->xrs);

    cairo_translate (win->xrs,
		     win->x_flip ? win->width - win->tx : win->tx,
		     win->y_flip ? win->height - win->ty : win->ty);
    cairo_scale (win->xrs,
		 win->x_flip ? -win->x_zoom : win->x_zoom,
		 win->y_flip ? -win->y_zoom : win->y_zoom);
    cairo_set_tolerance (win->xrs, win->tolerance);

/*
    start_traps = cairo_trapezoid_count (win->xrs);
    gettimeofday (&start_tv, NULL);
*/

    xsvg_render (win->xsvg, win->xrs);

/*
    gettimeofday (&end_tv, NULL);
    end_traps = cairo_trapezoid_count (win->xrs);

    elapsed = (end_tv.tv_sec - start_tv.tv_sec
	       + (end_tv.tv_usec - start_tv.tv_usec) / 1000000.0);
    fprintf (stderr, "Rendered %g trapezoids in %g seconds, (%g traps/sec)\n",
	     end_traps - start_traps,
	     elapsed,
	     (end_traps - start_traps) / elapsed);
*/

    cairo_restore (win->xrs);

    XCopyArea (win->dpy, win->pix, win->win, win->gc,
	       0, 0, win->width, win->height,
	       0, 0);
}

static void
win_grow_pixmap(win_t *win)
{
    Pixmap new;

    new = XCreatePixmap(win->dpy, win->win, win->width, win->height,
			DefaultDepth(win->dpy, win->scr));
    XFillRectangle(win->dpy, new, win->gc, 0, 0, win->width, win->height);
    XCopyArea(win->dpy, win->pix, new, win->gc, 0, 0, win->width, win->height, 0, 0);
    XFreePixmap(win->dpy, win->pix);
    win->pix = new;
    cairo_set_target_drawable (win->xrs, win->dpy, win->pix);
    win->needs_refresh = 1;
}

static int
win_handle_key_press(win_t *win, XKeyEvent *kev)
{
    unsigned int i;

    for (i=0; i < ARRAY_SIZE(key_binding); i++)
	if (key_binding[i].keycode == kev->keycode)
	    return (key_binding[i].callback)(win);
	
    return 0;
}

static void
win_reconfigure (win_t *win, unsigned int width, unsigned int height)
{
    int has_grown = 0;

    if (width > win->width || height > win->height)
	has_grown = 1;

    win->width = width;
    win->height = height;

    xsvg_set_viewport_dimension (win->xsvg, win->width, win->height);

    if (has_grown)
	win_grow_pixmap (win);
}

static void
win_handle_configure (win_t *win, XConfigureEvent *cev)
{
    win_reconfigure (win, cev->width, cev->height);
}

static void
win_handle_expose(win_t *win, XExposeEvent *eev)
{
    XCopyArea(win->dpy, win->pix, win->win, win->gc,
	      eev->x, eev->y, eev->width, eev->height,
	      eev->x, eev->y);
}

static void
win_handle_events(win_t *win)
{
    int done;
    XEvent xev;

    while (1) {
	if (win->needs_refresh) {
	    win_refresh(win);
	    win->needs_refresh = 0;
	} else {
	    /* XXX: This is pretty cheesy */
	    usleep(100000);
	}

	while (XCheckMaskEvent(win->dpy, win->event_mask, &xev)) {

	    switch(xev.type) {
	    case KeyPress:
		done = win_handle_key_press(win, &xev.xkey);
		if (done)
		    return;
		break;
	    case ConfigureNotify:
		win_handle_configure(win, &xev.xconfigure);
		break;
	    case Expose:
		win_handle_expose(win, &xev.xexpose);
		break;
	    }
	}
    }
}

static int
quit_cb(win_t *win)
{
    return 1;
}

static int
x_flip_cb(win_t *win)
{
    win->x_flip = ! win->x_flip;

    win->needs_refresh = 1;

    return 0;
}

static int
y_flip_cb(win_t *win)
{
    win->y_flip = ! win->y_flip;

    win->needs_refresh = 1;

    return 0;
}

static int
left_cb(win_t *win)
{
    if (win->x_flip)
	win->tx += SHIFT * win->x_zoom;
    else
	win->tx -= SHIFT * win->x_zoom;

    win->needs_refresh = 1;

    return 0;
}

static int
right_cb(win_t *win)
{
    if (win->x_flip)
	win->tx -= SHIFT * win->x_zoom;
    else
	win->tx += SHIFT * win->x_zoom;

    win->needs_refresh = 1;

    return 0;
}

static int
up_cb(win_t *win)
{
    if (win->y_flip)
	win->ty += SHIFT * win->y_zoom;
    else
	win->ty -= SHIFT * win->y_zoom;

    win->needs_refresh = 1;

    return 0;
}

static int
down_cb(win_t *win)
{
    if (win->y_flip)
	win->ty -= SHIFT * win->y_zoom;
    else
	win->ty += SHIFT * win->y_zoom;

    win->needs_refresh = 1;

    return 0;
}

static int
zoom_in_cb(win_t *win)
{
    win->x_zoom *= 1.1;
    win->y_zoom *= 1.1;

    win->needs_refresh = 1;

    return 0;
}

static int
zoom_x_cb(win_t *win)
{
    win->x_zoom *= 1.1;

    win->needs_refresh = 1;

    return 0;
}

static int
zoom_y_cb(win_t *win)
{
    win->y_zoom *= 1.1;

    win->needs_refresh = 1;

    return 0;
}

static int
zoom_out_cb(win_t *win)
{
    win->x_zoom /= 1.1;
    win->y_zoom /= 1.1;

    win->needs_refresh = 1;

    return 0;
}

static int
flatten_cb(win_t *win)
{
    win->tolerance *= 10;

    win->needs_refresh = 1;

    return 0;
}

static int
smooth_cb(win_t *win)
{
    win->tolerance /= 10;

    win->needs_refresh = 1;

    return 0;
}

static void
unpremultiply_data (png_structp png, png_row_infop row_info, png_bytep data)
{
    int i;

    for (i = 0; i < row_info->rowbytes; i += 4) {
        unsigned char *b = &data[i];
        unsigned char alpha = b[3];
        unsigned long pixel;
        unsigned long *p;
        if (alpha == 0)
            pixel = 0;
        else
            pixel = ((((b[0] * 255) / alpha) << 0) |
                     (((b[1] * 255) / alpha) << 8) |
                     (((b[2] * 255) / alpha) << 16) |
                     (alpha << 24));
        p = (unsigned long *) b;
        *p = pixel;
    }
}

static svg_status_t
write_png_argb32 (char *buffer, const char* filename,
		  int width, int height, int stride)
{
    FILE* f;
    png_struct* png;
    png_info* info;
    png_byte** rows;
    int       i;
    png_color_16 white;
    
    f = fopen (filename, "w");
    rows = malloc (height*sizeof (png_byte*));

    for (i=0;i<height;i++) {
	rows[i]=buffer+i*stride;
    }

    png = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    info = png_create_info_struct (png);

    png_init_io (png, f);
    png_set_IHDR (png, info,
		  width, height, 8,
		  PNG_COLOR_TYPE_RGB_ALPHA, 
		  PNG_INTERLACE_NONE,
		  PNG_COMPRESSION_TYPE_DEFAULT,
		  PNG_FILTER_TYPE_DEFAULT);

    white.red=0xff;
    white.blue=0xff;
    white.green=0xff;
    png_set_bKGD (png, info, &white);

    png_set_write_user_transform_fn (png, unpremultiply_data);
    png_set_bgr (png);

    png_write_info (png, info);
    png_write_image (png, rows);
    png_write_end (png, info);

    png_destroy_write_struct (&png, &info);

    free (rows);
    fclose (f);

    return SVG_STATUS_SUCCESS;
}
